@@ -9,6 +9,7 @@ import pickle
 import absl.logging
 absl.logging.set_verbosity(absl.logging.ERROR)
 import pandas as pd
+import gc
 
 
 # Distribution Sampling functions
@@ -689,6 +690,39 @@ class CausalConvEncoder(tf.keras.layers.Layer):
         state = x #[:,-1:,:] # return last hidden state 
         return state
 
+
+class CausalConvResidualLayer(tf.keras.layers.Layer):
+    def __init__(self, d_model, seq_len, dropout_rate):
+        super(CausalConvResidualLayer, self).__init__()
+        self.seq_len = seq_len
+        num_causal_layers = int(min_power_of_2(self.seq_len))
+
+        self.conv1x1 = tf.keras.layers.Conv1D(filters=d_model, kernel_size=1)
+        self.causalconvlayer1 = CausalConvEncoder(d_model=d_model, num_layers=num_causal_layers)
+
+        self.conv_dropout1 = tf.keras.layers.Dropout(dropout_rate)
+
+        self.conv_layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
+
+        self.conv_activation1 = tf.keras.layers.Activation('selu')
+
+        self.conv_add = tf.keras.layers.Add()
+
+    def call(self, inputs, training):
+        x = inputs
+
+        y = self.conv1x1(x)
+
+        x = self.causalconvlayer1(x)
+        x = self.conv_layernorm1(x)
+        x = self.conv_activation1(x)
+        x = self.conv_dropout1(x)
+
+        out = self.conv_add([x, y])
+
+        return out
+
+
 # Variable Weighted Transformer Model
 
 class SageTransformer(tf.keras.Model):
@@ -725,6 +759,8 @@ class SageTransformer(tf.keras.Model):
         
         if self.loss_fn == 'Point':
             self.final_layer = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=1))
+        elif self.loss_fn == 'Binary':
+            self.final_layer = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=1, activation='sigmoid'))
         elif self.loss_fn == 'Poisson':
             self.final_layer = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=1, activation='softplus'))
         elif self.loss_fn == 'Quantile':
@@ -755,6 +791,7 @@ class SageTransformer(tf.keras.Model):
         
         # scale -- old
         #scale = scale[:,-1:,:]
+
         # scale -- new
         scale = scale[:, -1:, :]
         s_dim = scale.shape.as_list()[-1]
@@ -792,6 +829,8 @@ class SageTransformer(tf.keras.Model):
         
         if self.loss_fn == 'Point':
             out = self.final_layer(dec_output, training=training)
+        elif self.loss_fn == 'Binary':
+            out = self.final_layer(dec_output, training=training)
         elif self.loss_fn == 'Poisson':
             if scaler == 'mean_scaler':
                 mean = self.final_layer(dec_output, training=training) * scale
@@ -933,9 +972,10 @@ class SageTransformer_Model(tf.keras.Model):
         # Create Numerical Embedding (Linear Transform) Layers
         
         self.target_linear_transform_layer = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=d_model, use_bias=False)) 
-        self.scale_linear_transform_layer = tf.keras.layers.Dense(units=d_model, use_bias=False)
-        self.hist_encode_layer = CausalConvEncoder(d_model=d_model, num_layers=self.num_causal_layers) 
-      
+        #self.scale_linear_transform_layer = tf.keras.layers.Dense(units=d_model, use_bias=False)
+        self.hist_encode_layer = CausalConvEncoder(d_model=d_model, num_layers=self.num_causal_layers)
+        #self.hist_encode_layer = CausalConvResidualLayer(d_model=d_model, seq_len=int(self.hist_len+self.f_len), dropout_rate=dropout_rate)
+
         if len(self.stat_num_col_names)>0:
             self.stat_linear_transform_layers = {}
             for colname in self.stat_num_col_names:
@@ -1027,30 +1067,35 @@ class SageTransformer_Model(tf.keras.Model):
             # rel_age
             rel_age = tf.strings.to_number(inputs[:, :-1, -4:-3], out_type=tf.dtypes.float32)
             rel_age_dec = self.known_linear_transform_layers['rel_age'](rel_age)
+
             # append
             decoder_vars_list.append(rel_age_dec)
+
             # scale
-            scale = tf.strings.to_number(inputs[:, :-1, -3:-1], out_type=tf.dtypes.float32)
-            scale_log = tf.math.log(tf.math.sqrt(tf.abs(scale[:,:,0:1]))) # mean
-            scale_log = self.scale_linear_transform_layer(scale_log[:,-1,:])
+            scale = tf.strings.to_number(inputs[:,:-1, -3:-1], out_type=tf.dtypes.float32)
+            #scale_log = scale[:,:,0:1] #tf.math.log(tf.math.sqrt(tf.abs(scale[:,:,0:1]))) # mean
+            #scale_log = self.scale_linear_transform_layer(scale_log[:,-1,:])
             # append
-            static_vars_list.append(scale_log)
+            #static_vars_list.append(scale_log)
+
         elif r_dim == 1:
             # print(" mean r_dim")
             # rel_age
             rel_age = tf.strings.to_number(inputs[:, :-1, -3:-2], out_type=tf.dtypes.float32)
             rel_age_dec = self.known_linear_transform_layers['rel_age'](rel_age)
+
             # append
             decoder_vars_list.append(rel_age_dec)
+
             # scale
             scale = tf.strings.to_number(inputs[:, :-1, -2:-1], out_type=tf.dtypes.float32)
-            scale_log = tf.math.log(tf.math.sqrt(scale))
-            scale_log = self.scale_linear_transform_layer(scale_log[:,-1,:])
+            #scale_log = scale #tf.math.log(tf.math.sqrt(scale))
+            #scale_log = self.scale_linear_transform_layer(scale_log[:,-1,:])
             # append
-            static_vars_list.append(scale_log)
+            #static_vars_list.append(scale_log)
         
         # Append additional columns
-        stat_cols_ordered_list = stat_cols_ordered_list + ['scale']
+        #stat_cols_ordered_list = stat_cols_ordered_list + ['scale']
         future_cols_ordered_list = future_cols_ordered_list + ['rel_age']
         
         # mask
@@ -1122,7 +1167,7 @@ def SageTransformer_Train(model,
                         loss = loss_function(y_train * s_std + s_mean, [s, wts])
                     else:
                         loss = loss_function(y_train * s_std + s_mean, s)
-            elif loss_type in ['Point']:
+            elif loss_type in ['Point','Binary']:
                 if weighted_training:                               
                     loss = loss_function(y_train, [o, wts])
                 else:
@@ -1156,7 +1201,7 @@ def SageTransformer_Train(model,
                     loss = loss_function(y_test * s_std + s_mean, [s, wts])
                 else:
                     loss = loss_function(y_test * s_std + s_mean, s)
-        elif loss_type in ['Point']:
+        elif loss_type in ['Point','Binary']:
             if weighted_training:                               
                 loss = loss_function(y_test, [o, wts])
             else:
@@ -1494,7 +1539,7 @@ def SageTransformer_InferRecursive(model, inputs, loss_type, hist_len, f_len, ta
                 infer_arr[:, hist_len:hist_len + i + 1, target_index] = out[:, 0:i + 1, 0] / scale
             else:
                 infer_arr[:, hist_len:hist_len + i + 1, target_index] = np.nan_to_num((out[:, 0:i + 1, 0] - scale_mean) / scale_std)
-        elif loss_type in ['Point','Quantile']:
+        elif loss_type in ['Point','Quantile','Binary']:
             out = out.numpy()
             output.append(out[:,i:i+1,0])
             infer_arr = infer_tensor.numpy()
@@ -1527,7 +1572,7 @@ def SageTransformer_InferRecursive(model, inputs, loss_type, hist_len, f_len, ta
     # rescale if necessary
     if loss_type in ['Normal','Poisson','Negbin']:
         output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1,1),output_arr), axis=1))
-    elif loss_type in ['Point','Quantile']:
+    elif loss_type in ['Point','Quantile','Binary']:
         #output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1,1),output_arr*scale.reshape(-1,1)), axis=1))
         # new
         if s_dim == 1:
