@@ -13,10 +13,13 @@ import shutil
 import absl.logging
 absl.logging.set_verbosity(absl.logging.ERROR)
 import pandas as pd
-import pickle
+
+
+# In[ ]:
 
 
 # Distribution Sampling functions
+
 # negbin
 def negbin_sample(mu, alpha, n_samples=1):
     tol = 1e-5
@@ -62,6 +65,8 @@ def GumbelSample(a, b, n_samples=1):
     return tf.reduce_mean(tf.stop_gradient(dist.sample(sample_shape=n_samples)), axis=0)
 
 
+# In[ ]:
+
 
 # masks
 def create_causal_mask(size):
@@ -179,6 +184,11 @@ class LocalSparseMultiHeadAttention(tf.keras.layers.Layer):
         output = self.dense(concat_attention, training=training)  # (batch_size, seq_len_q, d_model)
         
         return output, attention_weights
+      
+
+
+# In[ ]:
+
 
 # Encoder & Decoder Layers
 
@@ -290,6 +300,10 @@ class Decoder(tf.keras.layers.Layer):
         # x.shape == (batch_size, target_seq_len, d_model)
         return x, attention_weights
 
+
+# In[ ]:
+
+
 # variable selection networks
 
 class add_norm_layer(tf.keras.layers.Layer):
@@ -504,6 +518,11 @@ class temporal_variable_selection_layer(tf.keras.layers.Layer):
         tfr_input = tf.reduce_sum(combined, axis=2) #[batch,time_steps,hidden_layers_size]
         
         return tfr_input, dynamic_weights
+      
+
+
+# In[ ]:
+
 
 # Variable Weighted Transformer Model
 
@@ -627,6 +646,9 @@ class SparseVarTransformer(tf.keras.Model):
             return out, parameters, static_weights, encoder_weights, decoder_weights
 
 
+# In[ ]:
+
+
 # SparseVarTransformer Wrapper
 
 class SparseVarTransformer_Model(tf.keras.Model):
@@ -649,15 +671,6 @@ class SparseVarTransformer_Model(tf.keras.Model):
         self.f_len = int(forecast_horizon)
         self.loss_type = loss_type
         self.col_index_dict = col_index_dict
-        self.vocab_dict = vocab_dict
-        self.num_layers = num_layers
-        self.num_heads = num_heads
-        self.num_blocks = num_blocks
-        self.kernel_size = kernel_size
-        self.d_model = d_model
-        self.forecast_horizon = forecast_horizon
-        self.max_inp_len = max_inp_len
-        self.dropout_rate = dropout_rate
         self.decoder_lags = max(int(self.hist_len/4), 2)
         
         if (len(self.col_index_dict.get('static_num_indices')[0])==0) and (len(self.col_index_dict.get('static_cat_indices')[0])==0):
@@ -994,32 +1007,6 @@ def SparseVarTransformer_Train(model,
     test_metric_results = []
         
     # initialize model tracking vars
-    columns_dict_file = model_prefix + '_col_index_dict.pkl'
-    with open(columns_dict_file, 'wb') as f:
-        pickle.dump(model.col_index_dict, f)
-
-    vocab_dict_file = model_prefix + '_vocab_dict.pkl'
-    with open(vocab_dict_file, 'wb') as f:
-        pickle.dump(model.vocab_dict, f)
-    
-    model_tracker_file = open(model_prefix + '_tracker.txt', mode='w', encoding='utf-8')
-
-    model_tracker_file.write('Sparse Feature Weighted Transformer Training started with following Model Parameters ... \n')
-    model_tracker_file.write('----------------------------------------\n')
-    model_tracker_file.write('num_layers ' + str(model.num_layers) + '\n')
-    model_tracker_file.write('num_heads ' + str(model.num_heads) + '\n')
-    model_tracker_file.write('d_model ' + str(model.d_model) + '\n')
-    model_tracker_file.write('num_blocks ' + str(model.num_blocks) + '\n')
-    model_tracker_file.write('kernel_size ' + str(model.kernel_size) + '\n')
-    model_tracker_file.write('forecast_horizon ' + str(model.forecast_horizon) + '\n')
-    model_tracker_file.write('max_inp_len ' + str(model.max_inp_len) + '\n')
-    model_tracker_file.write('dropout_rate ' + str(model.dropout_rate) + '\n')
-    model_tracker_file.write('col_index_dict path ' + str(columns_dict_file) + '\n')
-    model_tracker_file.write('vocab_dict path ' + str(vocab_dict_file) + '\n')
-    model_tracker_file.write('----------------------------------------\n')
-    model_tracker_file.write('\n')
-    model_tracker_file.flush()
-    
     model_list = []
     best_model = None
     time_since_improvement = 0
@@ -1077,20 +1064,17 @@ def SparseVarTransformer_Train(model,
         
         # Save Model
         model_path = model_prefix + '_' + str(epoch) 
+        tf.keras.models.save_model(model, model_path)
         model_list.append(model_path)
             
         # track best model
         if test_loss_results[epoch]==np.min(test_loss_results):
             best_model = model_path
-            tf.keras.models.save_model(model, model_path)
             # reset time_since_improvement
             time_since_improvement = 0
         else:
             time_since_improvement += 1
-        
-        model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
-        print("Best Model: ", best_model)
-        
+            
         # remove older models
         if len(model_list)>patience:
             for m in model_list[:-patience]:
@@ -1099,14 +1083,11 @@ def SparseVarTransformer_Train(model,
                         shutil.rmtree(m)
                     except:
                         pass
-       
+                
+        print("Best Model: ", best_model)
         if (time_since_improvement > patience) and (epoch > min_epochs):
             print("Terminating Training. Best Model path: {}".format(best_model))
-            model_tracker_file.close()
             break
-            
-        # flush each epoch    
-        model_tracker_file.flush()
     
     return best_model
         
@@ -1185,6 +1166,10 @@ def SparseVarTransformer_Infer(model, inputs, loss_type, hist_len, f_len, target
         
     return forecast_df, stat_wts_df, encoder_wts_df, decoder_wts_df
 
+
+# In[ ]:
+
+
 class Sparse_Feature_Weighted_Transformer:
     def __init__(self, 
                  col_index_dict,
@@ -1285,6 +1270,11 @@ class Sparse_Feature_Weighted_Transformer:
         results_df['Forecast_Bias'] = (results_df['forecast']/np.maximum(results_df['actual'],1.0) - 1)*100
         
         return results_df
+    
+
+
+# In[ ]:
+
 
 # Transformer Base Model
 
@@ -1378,6 +1368,10 @@ class SparseTransformer(tf.keras.Model):
         else:
             return out, parameters
 
+
+# In[ ]:
+
+
 # SparseTransformer Wrapper
 
 class SparseTransformer_Model(tf.keras.Model):
@@ -1414,15 +1408,6 @@ class SparseTransformer_Model(tf.keras.Model):
         self.f_len = int(forecast_horizon)
         self.loss_type = loss_type
         self.col_index_dict = col_index_dict
-        self.vocab_dict = vocab_dict
-        self.num_layers = num_layers
-        self.num_heads = num_heads
-        self.num_blocks = num_blocks
-        self.kernel_size = kernel_size
-        self.d_model = d_model
-        self.forecast_horizon = forecast_horizon
-        self.max_inp_len = max_inp_len
-        self.dropout_rate = dropout_rate
         self.decoder_lags = max(int(self.hist_len/4), 2)
         
         self.model = SparseTransformer(num_layers=num_layers,
@@ -1699,32 +1684,6 @@ def SparseTransformer_Train(model,
     test_metric_results = []
         
     # initialize model tracking vars
-    columns_dict_file = model_prefix + '_col_index_dict.pkl'
-    with open(columns_dict_file, 'wb') as f:
-        pickle.dump(model.col_index_dict, f)
-
-    vocab_dict_file = model_prefix + '_vocab_dict.pkl'
-    with open(vocab_dict_file, 'wb') as f:
-        pickle.dump(model.vocab_dict, f)
-    
-    model_tracker_file = open(model_prefix + '_tracker.txt', mode='w', encoding='utf-8')
-
-    model_tracker_file.write('Sparse Simple Transformer Training started with following Model Parameters ... \n')
-    model_tracker_file.write('----------------------------------------\n')
-    model_tracker_file.write('num_layers ' + str(model.num_layers) + '\n')
-    model_tracker_file.write('num_heads ' + str(model.num_heads) + '\n')
-    model_tracker_file.write('d_model ' + str(model.d_model) + '\n')
-    model_tracker_file.write('num_blocks ' + str(model.num_blocks) + '\n')
-    model_tracker_file.write('kernel_size ' + str(model.kernel_size) + '\n')
-    model_tracker_file.write('forecast_horizon ' + str(model.forecast_horizon) + '\n')
-    model_tracker_file.write('max_inp_len ' + str(model.max_inp_len) + '\n')
-    model_tracker_file.write('dropout_rate ' + str(model.dropout_rate) + '\n')
-    model_tracker_file.write('col_index_dict path ' + str(columns_dict_file) + '\n')
-    model_tracker_file.write('vocab_dict path ' + str(vocab_dict_file) + '\n')
-    model_tracker_file.write('----------------------------------------\n')
-    model_tracker_file.write('\n')
-    model_tracker_file.flush()
-    
     model_list = []
     best_model = None
     time_since_improvement = 0
@@ -1782,20 +1741,17 @@ def SparseTransformer_Train(model,
         
         # Save Model
         model_path = model_prefix + '_' + str(epoch) 
+        tf.keras.models.save_model(model, model_path)
         model_list.append(model_path)
             
         # track best model
         if test_loss_results[epoch]==np.min(test_loss_results):
             best_model = model_path
-            tf.keras.models.save_model(model, model_path)
             # reset time_since_improvement
             time_since_improvement = 0
         else:
             time_since_improvement += 1
-        
-        model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
-        print("Best Model: ", best_model)
-        
+            
         # remove older models
         if len(model_list)>patience:
             for m in model_list[:-patience]:
@@ -1805,12 +1761,11 @@ def SparseTransformer_Train(model,
                     except:
                         pass
                 
+        print("Best Model: ", best_model)
         if (time_since_improvement > patience) and (epoch > min_epochs):
             print("Terminating Training. Best Model path: {}".format(best_model))
-            model_tracker_file.close()
             break
-        # flush after each epoch    
-        model_tracker_file.flush()
+    
     return best_model
     
     
@@ -1857,6 +1812,9 @@ def SparseTransformer_Infer(model, inputs, loss_type, hist_len, f_len, target_in
     return forecast_df
 
 
+# In[ ]:
+
+
 class Sparse_Simple_Transformer:
     def __init__(self, 
                  col_index_dict,
@@ -1958,3 +1916,5 @@ class Sparse_Simple_Transformer:
         
         return results_df
         
+                
+
