@@ -14,6 +14,7 @@ import pickle
 import absl.logging
 absl.logging.set_verbosity(absl.logging.ERROR)
 import pandas as pd
+import gc
 
 # Distribution Sampling functions
 
@@ -666,6 +667,21 @@ class TFT(tf.keras.Model):
         if self.context:
             stat_vars_list, encoder_vars_list, decoder_vars_list, padding_mask, scale = inputs
 
+            # scale
+            scale = scale[:, -1:, :]
+            s_dim = scale.shape.as_list()[-1]
+
+            if s_dim == 2:
+                # print("standard scaling used")
+                scaler = 'standard_scaler'
+                scale_mean = scale[:, :, 0:1]
+                scale_std = scale[:, :, 1:2]
+            else:
+                # print("mean scaling used")
+                scaler = 'mean_scaler'
+                scale_mean = scale[:, :, 0:1]
+                scale_std = scale[:, :, 0:1]
+
             static_vec, stat_weights = self.static_var_select_layer(stat_vars_list, training=training)
             context_vec, enrichment_vec, init_h, init_c = self.static_context_layer(static_vec, training=training)
             init_states = [init_h, init_c]
@@ -674,6 +690,22 @@ class TFT(tf.keras.Model):
             decoder_inputs, future_weights = self.temporal_var_select_decoder_layer([decoder_vars_list, context_vec], training=training)
         else:
             encoder_vars_list, decoder_vars_list, padding_mask, scale = inputs
+
+            # scale
+            scale = scale[:, -1:, :]
+            s_dim = scale.shape.as_list()[-1]
+
+            if s_dim == 2:
+                # print("standard scaling used")
+                scaler = 'standard_scaler'
+                scale_mean = scale[:, :, 0:1]
+                scale_std = scale[:, :, 1:2]
+            else:
+                # print("mean scaling used")
+                scaler = 'mean_scaler'
+                scale_mean = scale[:, :, 0:1]
+                scale_std = scale[:, :, 0:1]
+
             batch_size = tf.shape(encoder_vars_list[0])[0]
             init_h = tf.zeros([batch_size, self.hidden_layer_size], dtype=tf.float32)
             init_c = tf.zeros([batch_size, self.hidden_layer_size], dtype=tf.float32)
@@ -682,6 +714,7 @@ class TFT(tf.keras.Model):
             encoder_inputs, past_weights = self.temporal_var_select_encoder_layer(encoder_vars_list, training=training)
             decoder_inputs, future_weights = self.temporal_var_select_decoder_layer(decoder_vars_list, training=training)
 
+
         # recurrent layer
         temporal_features = self.recurrent_layer([encoder_inputs, decoder_inputs, init_states], training=training)
 
@@ -708,17 +741,33 @@ class TFT(tf.keras.Model):
         elif self.loss_fn == 'Binary':
             out = self.final_layer(attn_out, training=training)
         elif self.loss_fn == 'Poisson':
-            mean = self.final_layer(attn_out, training=training)*scale
+            if scaler == 'mean_scaler':
+                mean = self.final_layer(attn_out, training=training) * scale
+            elif scaler == 'standard_scaler':
+                mean = self.final_layer(attn_out, training=training) * scale_std + scale_mean
+            #mean = self.final_layer(attn_out, training=training)*scale
             parameters = mean
             out = poisson_sample(mu=mean)
         elif self.loss_fn == 'Normal':
-            mean = self.m_layer(attn_out, training=training)*scale
-            stddev = self.s_layer(attn_out, training=training)*scale
+            if scaler == 'mean_scaler':
+                mean = self.m_layer(attn_out, training=training) * scale  # (batch, f_len, 1)
+                stddev = self.s_layer(attn_out, training=training) * scale  # (batch, f_len, 1)
+            elif scaler == 'standard_scaler':
+                mean = self.m_layer(attn_out, training=training) * scale_std + scale_mean  # (batch, f_len, 1)
+                stddev = self.s_layer(attn_out, training=training) * scale_std  # (batch, f_len, 1)
+            #mean = self.m_layer(attn_out, training=training)*scale
+            #stddev = self.s_layer(attn_out, training=training)*scale
             parameters = tf.concat([mean, stddev], axis=-1)
             out = normal_sample(mean, stddev)
         elif self.loss_fn == 'Negbin':
-            mean = self.m_layer(attn_out, training=training)*scale
-            alpha = self.a_layer(attn_out, training=training)*tf.sqrt(scale)
+            if scaler == 'mean_scaler':
+                mean = self.m_layer(attn_out, training=training) * scale  # (batch, f_len, 1)
+                alpha = self.a_layer(attn_out, training=training) * tf.sqrt(scale)  # (batch, f_len, 1)
+            elif scaler == 'standard_scaler':
+                mean = self.m_layer(attn_out, training=training) * scale_std + scale_mean  # (batch, f_len, 1)
+                alpha = self.a_layer(attn_out, training=training) * tf.sqrt(scale_std)  # (batch, f_len, 1)
+            #mean = self.m_layer(attn_out, training=training)*scale
+            #alpha = self.a_layer(attn_out, training=training)*tf.sqrt(scale)
             parameters = tf.concat([mean, alpha], axis=-1)
             out = negbin_sample(mean, alpha)
         elif self.loss_fn == 'Quantile':
@@ -853,7 +902,7 @@ class TFT_Model(tf.keras.Model):
         # Create Numerical Embedding (Linear Transform) Layers
         
         self.target_linear_transform_layer = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=d_model, use_bias=False)) 
-        self.scale_linear_transform_layer = tf.keras.layers.Dense(units=d_model, use_bias=False)
+        #self.scale_linear_transform_layer = tf.keras.layers.Dense(units=d_model, use_bias=False)
         
         if len(self.stat_num_col_names)>0:
             self.stat_linear_transform_layers = {}
@@ -867,14 +916,19 @@ class TFT_Model(tf.keras.Model):
         
         if len(self.unknown_num_col_names)>=0:
             self.unknown_linear_transform_layers = {}
-            for colname in self.unknown_num_col_names + ['rel_age']:
+            for colname in self.unknown_num_col_names:
                 self.unknown_linear_transform_layers[colname] = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=d_model, use_bias=False))
                 
     def call(self, inputs, training):
+
+        # total_dim
+        t_dim = inputs.shape.as_list()[-1] - 3  # reduce 2 dims for mask,rel_age,scale
+        dim_counter = 0
         
         # target
         target = tf.strings.to_number(inputs[:,:,self.target_index:self.target_index+1], out_type=tf.dtypes.float32)
         target = self.target_linear_transform_layer(target)
+        dim_counter += 1
         
         # ordered col names list
         stat_cols_ordered_list = []
@@ -903,6 +957,7 @@ class TFT_Model(tf.keras.Model):
                 stat_var = self.stat_linear_transform_layers[col](stat_var)
                 # append
                 static_vars_list.append(stat_var)
+                dim_counter += 1
         
         # static embeddings
         if len(self.stat_cat_indices)>0:
@@ -913,6 +968,7 @@ class TFT_Model(tf.keras.Model):
                 stat_var_embeddings = self.stat_embed_layers.get(col)(stat_var_id)
                 # append
                 static_vars_list.append(stat_var_embeddings[:,-1,:])
+                dim_counter += 1
                 
         # known numeric 
         if len(self.known_num_indices)>0:
@@ -924,6 +980,7 @@ class TFT_Model(tf.keras.Model):
                 # append
                 encoder_vars_list.append(num_vars[:,:self.hist_len,:])
                 decoder_vars_list.append(num_vars[:,-self.f_len:,:])
+                dim_counter += 1
         
         # unknown numeric
         if len(self.unknown_num_indices)>0:
@@ -933,6 +990,7 @@ class TFT_Model(tf.keras.Model):
                 num_vars = self.unknown_linear_transform_layers[col](num_vars)
                 # append
                 encoder_vars_list.append(num_vars[:,:self.hist_len,:])
+                dim_counter += 1
                      
         # known embeddings
         if len(self.known_cat_indices)>0:
@@ -945,6 +1003,7 @@ class TFT_Model(tf.keras.Model):
                 # append
                 encoder_vars_list.append(cat_var_embeddings[:,:self.hist_len,:])
                 decoder_vars_list.append(cat_var_embeddings[:,-self.f_len:,:])
+                dim_counter += 1
         
         # unknown embeddings
         if len(self.unknown_cat_indices)>0:
@@ -955,7 +1014,17 @@ class TFT_Model(tf.keras.Model):
                 cat_var_embeddings = self.temporal_unknown_embed_layers.get(col)(cat_var_id)
                 # append
                 encoder_vars_list.append(cat_var_embeddings[:,:self.hist_len,:])
-                          
+                dim_counter += 1
+
+        # remaining_dim
+        r_dim = t_dim - dim_counter
+
+        # default scale
+        scale = tf.strings.to_number(inputs[:, :, -2:-1], out_type=tf.dtypes.float32)
+
+        '''   
+        # original
+                       
         # rel_age
         rel_age = tf.strings.to_number(inputs[:,:,-3:-2], out_type=tf.dtypes.float32)
         rel_age_enc = self.known_linear_transform_layers['rel_age'](rel_age[:,:self.hist_len,:])
@@ -970,14 +1039,37 @@ class TFT_Model(tf.keras.Model):
         scale_log = self.scale_linear_transform_layer(scale_log[:,-1,:])
         # append
         static_vars_list.append(scale_log)
-        
+        '''
+
+        if r_dim == 2:  # standard scaling used (mean,std)
+            # print(" standard r_dim")
+            # rel_age
+            rel_age = tf.strings.to_number(inputs[:, :, -4:-3], out_type=tf.dtypes.float32)
+            rel_age_enc = self.known_linear_transform_layers['rel_age'](rel_age)
+            # append
+            encoder_vars_list.append(rel_age_enc[:, :self.hist_len, :])
+            decoder_vars_list.append(rel_age_enc[:, -self.f_len:, :])
+
+            # scale
+            scale = tf.strings.to_number(inputs[:, :, -3:-1], out_type=tf.dtypes.float32)
+
+        elif r_dim == 1:
+            # print(" mean r_dim")
+            # rel_age
+            rel_age = tf.strings.to_number(inputs[:, :, -3:-2], out_type=tf.dtypes.float32)
+            rel_age_enc = self.known_linear_transform_layers['rel_age'](rel_age)
+            # append
+            encoder_vars_list.append(rel_age_enc[:, :self.hist_len, :])
+            decoder_vars_list.append(rel_age_enc[:, -self.f_len:, :])
+            # scale
+            scale = tf.strings.to_number(inputs[:, :, -2:-1], out_type=tf.dtypes.float32)
+
         # Append additional columns
-        stat_cols_ordered_list = stat_cols_ordered_list + ['scale']
         past_cols_ordered_list = past_cols_ordered_list + ['rel_age']
         future_cols_ordered_list = future_cols_ordered_list + ['rel_age']
         
         # mask
-        mask = tf.strings.to_number(inputs[:,:,-1:], out_type=tf.dtypes.float32)
+        mask = tf.strings.to_number(inputs[:, :, -1:], out_type=tf.dtypes.float32)
         
         # model process
         if self.static_variables:
@@ -986,8 +1078,13 @@ class TFT_Model(tf.keras.Model):
             o, s, s_wts, p_wts, f_wts = self.model([encoder_vars_list, decoder_vars_list, mask, scale], training=training)
         
         # Average feature weights across time dim
-        p_wts = tf.math.reduce_mean(p_wts, axis=1)
-        f_wts = tf.math.reduce_mean(f_wts, axis=1)
+        #p_wts = tf.math.reduce_mean(p_wts, axis=1)
+        #f_wts = tf.math.reduce_mean(f_wts, axis=1)
+
+        # Retain period-wise importance
+        bs = tf.shape(p_wts)[0]
+        p_wts = tf.reshape(p_wts, [bs * self.hist_len, -1])
+        f_wts = tf.reshape(f_wts, [bs * self.f_len, -1])
         
         return o, s, ([stat_cols_ordered_list,past_cols_ordered_list,future_cols_ordered_list], [s_wts, p_wts, f_wts])
     
@@ -1001,6 +1098,10 @@ def TFT_Train(model,
               learning_rate,
               max_epochs, 
               min_epochs,
+              prefill_buffers,
+              num_train_samples,
+              num_test_samples,
+              train_batch_size,
               train_steps_per_epoch,
               test_steps_per_epoch,
               patience,
@@ -1026,13 +1127,28 @@ def TFT_Train(model,
     def trainstep(model, optimizer, x_train, y_train, scale, wts, training):
         with tf.GradientTape() as tape:
             o, s, f = model(x_train, training=training)
-            out_len = tf.shape(s)[1]
+            out_len = s.shape.as_list()[1]  # tf.shape(s)[1]
+            s_dim = scale.shape.as_list()[-1]  # tf.shape(scale)[-1]
             if loss_type in ['Normal','Poisson','Negbin']:
+                '''
                 if weighted_training:
                     loss = loss_function(y_train*scale[:,-out_len:,:], [s, wts])
                 else:
-                    loss = loss_function(y_train*scale[:,-out_len:,:], s)                           
-            elif loss_type in ['Point']:
+                    loss = loss_function(y_train*scale[:,-out_len:,:], s)
+                '''
+                if s_dim == 1:
+                    if weighted_training:
+                        loss = loss_function(y_train * scale[:, -out_len:, :], [s, wts])
+                    else:
+                        loss = loss_function(y_train * scale[:, -out_len:, :], s)
+                else:
+                    s_mean = scale[:, -out_len:, 0:1]
+                    s_std = scale[:, -out_len:, 1:2]
+                    if weighted_training:
+                        loss = loss_function(y_train * s_std + s_mean, [s, wts])
+                    else:
+                        loss = loss_function(y_train * s_std + s_mean, s)
+            elif loss_type in ['Point','Binary']:
                 if weighted_training:                               
                     loss = loss_function(y_train, [o, wts])
                 else:
@@ -1051,13 +1167,28 @@ def TFT_Train(model,
     @tf.function
     def teststep(model, x_test, y_test, scale, wts, training):
         o, s, f = model(x_test, training=training)
-        out_len = tf.shape(s)[1]
+        out_len = s.shape.as_list()[1]  # tf.shape(s)[1]
+        s_dim = scale.shape.as_list()[-1]  # tf.shape(scale)[-1]
         if loss_type in ['Normal','Poisson','Negbin']:
+            '''
             if weighted_training:
                 loss = loss_function(y_test*scale[:,-out_len:,:], [s, wts])
             else:
-                loss = loss_function(y_test*scale[:,-out_len:,:], s)                           
-        elif loss_type in ['Point']:
+                loss = loss_function(y_test*scale[:,-out_len:,:], s)    
+            '''
+            if s_dim == 1:
+                if weighted_training:
+                    loss = loss_function(y_test * scale[:, -out_len:, :], [s, wts])
+                else:
+                    loss = loss_function(y_test * scale[:, -out_len:, :], s)
+            else:
+                s_mean = scale[:, -out_len:, 0:1]
+                s_std = scale[:, -out_len:, 1:2]
+                if weighted_training:
+                    loss = loss_function(y_test * s_std + s_mean, [s, wts])
+                else:
+                    loss = loss_function(y_test * s_std + s_mean, s)
+        elif loss_type in ['Point','Binary']:
             if weighted_training:                               
                 loss = loss_function(y_test, [o, wts])
             else:
@@ -1139,96 +1270,242 @@ def TFT_Train(model,
     model_list = []
     best_model = None
     time_since_improvement = 0
-        
-    # train loop
-    for epoch in range(max_epochs):
-        print("Epoch {}/{}". format(epoch, max_epochs)) 
+    #####################################################################
+
+    if prefill_buffers:
+        print("prefetching training samples ... ")
+        # get batch size
+        batch_size = 0
+        for x, y, s, w in train_dataset.take(1):
+            batch_size = int(x.shape[0])
+
+        x_train = []
+        y_train = []
+        train_scale = []
+        train_wts = []
         for step, (x_batch, y_batch, scale, wts) in enumerate(train_dataset):
-            if step > train_steps_per_epoch:
+            x_train.append(x_batch)
+            y_train.append(y_batch)
+            train_scale.append(scale)
+            train_wts.append(wts)
+            if (step + 1) * batch_size >= num_train_samples:
                 break
-            else:
+
+        # concat
+        x_train = tf.concat(x_train, axis=0)
+        y_train = tf.concat(y_train, axis=0)
+        train_scale = tf.concat(train_scale, axis=0)
+        train_wts = tf.concat(train_wts, axis=0)
+        print("Training Samples Gathered: ", x_train.shape[0])
+
+        print("prefetching test samples ... ")
+        x_test = []
+        y_test = []
+        test_scale = []
+        test_wts = []
+        for step, (x_batch, y_batch, scale, wts) in enumerate(test_dataset):
+            x_test.append(x_batch)
+            y_test.append(y_batch)
+            test_scale.append(scale)
+            test_wts.append(wts)
+            if (step + 1) * batch_size >= num_test_samples:
+                break
+
+        # concat
+        x_test = tf.concat(x_test, axis=0)
+        y_test = tf.concat(y_test, axis=0)
+        test_scale = tf.concat(test_scale, axis=0)
+        test_wts = tf.concat(test_wts, axis=0)
+        print("Test Samples Gathered: ", x_test.shape[0])
+
+        num_train_batches = int(x_train.shape[0] // train_batch_size)
+        num_test_batches = int(x_test.shape[0] // train_batch_size)
+
+        for epoch in range(max_epochs):
+            print("Epoch {}/{}".format(epoch, max_epochs))
+            for i in range(num_train_batches):
+                x_batch = x_train[i * train_batch_size:(i + 1) * train_batch_size]
+                y_batch = y_train[i * train_batch_size:(i + 1) * train_batch_size]
+                scale = train_scale[i * train_batch_size:(i + 1) * train_batch_size]
+                wts = train_wts[i * train_batch_size:(i + 1) * train_batch_size]
                 train_loss, train_out = trainstep(model, optimizer, x_batch, y_batch, scale, wts, training=True)
                 out_len = tf.shape(train_out)[1]
                 train_loss_avg.update_state(train_loss)
-                if loss_type in ['Normal','Poisson','Negbin']:
-                    train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out)
-                elif loss_type in ['Point','Quantile']:
-                    train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out*scale[:,-out_len:,:])
+                if loss_type in ['Normal', 'Poisson', 'Negbin']:
+                    train_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :], train_out)
+                elif loss_type in ['Point', 'Quantile']:
+                    train_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :],
+                                              train_out * scale[:, -out_len:, :])
                 with train_summary_writer.as_default():
                     tf.summary.scalar('loss', train_loss_avg.result(), step=epoch)
                     tf.summary.scalar('accuracy', train_metric.result(), step=epoch)
-          
-        for step, (x_batch, y_batch, scale, wts) in enumerate(test_dataset):
-            if step > train_steps_per_epoch:
-                break
-            else:
-                test_loss, test_out = teststep(model, x_batch, y_batch, scale, wts, training=False)
+
+            for i in range(num_test_batches):
+                x_batch = x_test[i * train_batch_size:(i + 1) * train_batch_size]
+                y_batch = y_test[i * train_batch_size:(i + 1) * train_batch_size]
+                scale = test_scale[i * train_batch_size:(i + 1) * train_batch_size]
+                wts = test_wts[i * train_batch_size:(i + 1) * train_batch_size]
+                test_loss, test_out = trainstep(model, optimizer, x_batch, y_batch, scale, wts, training=False)
                 out_len = tf.shape(test_out)[1]
                 test_loss_avg.update_state(test_loss)
-                if loss_type in ['Normal','Poisson','Negbin']:
-                    test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out)
-                elif loss_type in ['Point','Quantile']:
-                    test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out*scale[:,-out_len:,:])
+                if loss_type in ['Normal', 'Poisson', 'Negbin']:
+                    test_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :], test_out)
+                elif loss_type in ['Point', 'Quantile']:
+                    test_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :],
+                                             test_out * scale[:, -out_len:, :])
                 with test_summary_writer.as_default():
                     tf.summary.scalar('loss', test_loss_avg.result(), step=epoch)
                     tf.summary.scalar('accuracy', test_metric.result(), step=epoch)
-              
-        print("Epoch: {}, train_loss: {}, test_loss: {}, train_metric: {}, test_metric: {}".format(epoch, 
-                                                                                                    train_loss_avg.result().numpy(),
-                                                                                                    test_loss_avg.result().numpy(),
-                                                                                                    train_metric.result().numpy(),
-                                                                                                    test_metric.result().numpy()))
-    
-        # record losses & metric in lists
-        train_loss_results.append(train_loss_avg.result().numpy())
-        train_metric_results.append(train_metric.result().numpy())
-        test_loss_results.append(test_loss_avg.result().numpy())
-        test_metric_results.append(test_metric.result().numpy())
-            
-        # reset states
-        train_loss_avg.reset_states()
-        train_metric.reset_states()
-        test_loss_avg.reset_states()
-        test_metric.reset_states()
-        
-        # Save Model
-        model_path = model_prefix + '_' + str(epoch) 
-        model_list.append(model_path)
-        
-        # track & save best model
-        if test_loss_results[epoch]==np.min(test_loss_results):
-            best_model = model_path
-            tf.keras.models.save_model(model, model_path)
-            # reset time_since_improvement
-            time_since_improvement = 0
-        else:
-            time_since_improvement += 1
-            
-        model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
-        print("Best Model: ", best_model)
-            
-        # remove older models
-        if len(model_list)>patience:
-            for m in model_list[:-patience]:
-                if m != best_model:
-                    try:
-                        shutil.rmtree(m)
-                    except:
-                        pass
-                
-        if (time_since_improvement > patience) and (epoch > min_epochs):
-            print("Terminating Training. Best Model path: {}".format(best_model))
-            model_tracker_file.close()
-            break
-            
-        # write after each epoch
-        model_tracker_file.flush()
-        
+
+            print("Epoch: {}, train_loss: {}, test_loss: {}, train_metric: {}, test_metric: {}".format(epoch,
+                                                                                                       train_loss_avg.result().numpy(),
+                                                                                                       test_loss_avg.result().numpy(),
+                                                                                                       train_metric.result().numpy(),
+                                                                                                       test_metric.result().numpy()))
+
+            # record losses & metric in lists
+            train_loss_results.append(train_loss_avg.result().numpy())
+            train_metric_results.append(train_metric.result().numpy())
+            test_loss_results.append(test_loss_avg.result().numpy())
+            test_metric_results.append(test_metric.result().numpy())
+
+            # reset states
+            train_loss_avg.reset_states()
+            train_metric.reset_states()
+            test_loss_avg.reset_states()
+            test_metric.reset_states()
+
+            # Save Model
+            model_path = model_prefix + '_' + str(epoch)
+            model_list.append(model_path)
+
+            # track & save best model
+            if test_loss_results[epoch] == np.min(test_loss_results):
+                best_model = model_path
+                tf.keras.models.save_model(model, model_path)
+                # reset time_since_improvement
+                time_since_improvement = 0
+            else:
+                time_since_improvement += 1
+
+            model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
+            print("Best Model: ", best_model)
+
+            # remove older models
+            if len(model_list) > patience:
+                for m in model_list[:-patience]:
+                    if m != best_model:
+                        try:
+                            shutil.rmtree(m)
+                        except:
+                            pass
+
+            if (time_since_improvement > patience) and (epoch > min_epochs):
+                print("Terminating Training. Best Model path: {}".format(best_model))
+                model_tracker_file.close()
+                break
+
+            # write after each epoch
+            model_tracker_file.flush()
+    else:
+        # train loop
+        for epoch in range(max_epochs):
+            print("Epoch {}/{}". format(epoch, max_epochs))
+            for step, (x_batch, y_batch, scale, wts) in enumerate(train_dataset):
+                if step > train_steps_per_epoch:
+                    break
+                else:
+                    train_loss, train_out = trainstep(model, optimizer, x_batch, y_batch, scale, wts, training=True)
+                    out_len = tf.shape(train_out)[1]
+                    train_loss_avg.update_state(train_loss)
+                    if loss_type in ['Normal','Poisson','Negbin']:
+                        train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out)
+                    elif loss_type in ['Point','Quantile']:
+                        train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out*scale[:,-out_len:,:])
+                    with train_summary_writer.as_default():
+                        tf.summary.scalar('loss', train_loss_avg.result(), step=epoch)
+                        tf.summary.scalar('accuracy', train_metric.result(), step=epoch)
+
+            for step, (x_batch, y_batch, scale, wts) in enumerate(test_dataset):
+                if step > train_steps_per_epoch:
+                    break
+                else:
+                    test_loss, test_out = teststep(model, x_batch, y_batch, scale, wts, training=False)
+                    out_len = tf.shape(test_out)[1]
+                    test_loss_avg.update_state(test_loss)
+                    if loss_type in ['Normal','Poisson','Negbin']:
+                        test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out)
+                    elif loss_type in ['Point','Quantile']:
+                        test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out*scale[:,-out_len:,:])
+                    with test_summary_writer.as_default():
+                        tf.summary.scalar('loss', test_loss_avg.result(), step=epoch)
+                        tf.summary.scalar('accuracy', test_metric.result(), step=epoch)
+
+            print("Epoch: {}, train_loss: {}, test_loss: {}, train_metric: {}, test_metric: {}".format(epoch,
+                                                                                                        train_loss_avg.result().numpy(),
+                                                                                                        test_loss_avg.result().numpy(),
+                                                                                                        train_metric.result().numpy(),
+                                                                                                        test_metric.result().numpy()))
+
+            # record losses & metric in lists
+            train_loss_results.append(train_loss_avg.result().numpy())
+            train_metric_results.append(train_metric.result().numpy())
+            test_loss_results.append(test_loss_avg.result().numpy())
+            test_metric_results.append(test_metric.result().numpy())
+
+            # reset states
+            train_loss_avg.reset_states()
+            train_metric.reset_states()
+            test_loss_avg.reset_states()
+            test_metric.reset_states()
+
+            # Save Model
+            model_path = model_prefix + '_' + str(epoch)
+            model_list.append(model_path)
+
+            # track & save best model
+            if test_loss_results[epoch]==np.min(test_loss_results):
+                best_model = model_path
+                tf.keras.models.save_model(model, model_path)
+                # reset time_since_improvement
+                time_since_improvement = 0
+            else:
+                time_since_improvement += 1
+
+            model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
+            print("Best Model: ", best_model)
+
+            # remove older models
+            if len(model_list)>patience:
+                for m in model_list[:-patience]:
+                    if m != best_model:
+                        try:
+                            shutil.rmtree(m)
+                        except:
+                            pass
+
+            if (time_since_improvement > patience) and (epoch > min_epochs):
+                print("Terminating Training. Best Model path: {}".format(best_model))
+                model_tracker_file.close()
+                break
+
+            # write after each epoch
+            model_tracker_file.flush()
+
     return best_model
         
 def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quantiles):
     infer_tensor, scale, id_arr, date_arr = inputs
-    scale = scale[:,-1:,-1]
+
+    s_dim = scale.shape[-1]
+
+    if s_dim == 1:
+        scale = scale[:, -1:, -1]
+    else:
+        scale_mean = scale[:, -1:, 0]
+        scale_std = scale[:, -1:, 1]
+
+    #scale = scale[:,-1:,-1] #old
     window_len = hist_len + f_len
     stat_wts_df = None 
     encoder_wts_df = None 
@@ -1240,7 +1517,7 @@ def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quant
         dist = dist.numpy()
         output_arr = dist[:,:,0]
         
-    elif loss_type in ['Point']:
+    elif loss_type in ['Point','Binary']:
         out = out.numpy()
         output_arr = out[:,:,0]
     
@@ -1279,18 +1556,31 @@ def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quant
         output_df = output_df.melt(id_vars=0).sort_values(0).drop(columns=['variable']).rename(columns={0:'id','value':'forecast'})
         output_df = output_df.rename_axis('index').sort_values(by=['id','index']).reset_index(drop=True)
         
-    elif loss_type in ['Point']:
-        output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1,1),output_arr*scale.reshape(-1,1)), axis=1))
+    elif loss_type in ['Point','Binary']:
+
+        #output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1,1),output_arr*scale.reshape(-1,1)), axis=1))
+        if s_dim == 1:
+            output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1, 1), output_arr * scale.reshape(-1, 1)), axis=1))
+        else:
+            output_arr = output_arr * scale_std.reshape(-1, 1) + scale_mean.reshape(-1, 1)
+            output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1, 1), output_arr), axis=1))
         output_df = output_df.melt(id_vars=0).sort_values(0).drop(columns=['variable']).rename(columns={0:'id','value':'forecast'})
         output_df = output_df.rename_axis('index').sort_values(by=['id','index']).reset_index(drop=True)
         
     elif loss_type in ['Quantile']:
         df_list = []
         for i in range(num_quantiles):
-            output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1,1),output_arr[:,:,i]*scale.reshape(-1,1)), axis=1))
-            output_df = output_df.melt(id_vars=0).sort_values(0).drop(columns=['variable']).rename(columns={0:'id','value': f"forecast_{i}"})
-            output_df = output_df.rename_axis('index').sort_values(by=['id','index']).reset_index(drop=True)
-            df_list.append(output_df)
+            if s_dim == 1:
+                output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1,1),output_arr[:,:,i]*scale.reshape(-1,1)), axis=1))
+                output_df = output_df.melt(id_vars=0).sort_values(0).drop(columns=['variable']).rename(columns={0:'id','value': f"forecast_{i}"})
+                output_df = output_df.rename_axis('index').sort_values(by=['id','index']).reset_index(drop=True)
+                df_list.append(output_df)
+            else:
+                output_arr = output_arr[:,:,i]* scale_std.reshape(-1, 1) + scale_mean.reshape(-1, 1)
+                output_df = pd.DataFrame(np.concatenate((id_arr.reshape(-1, 1), output_arr), axis=1))
+                output_df = output_df.melt(id_vars=0).sort_values(0).drop(columns=['variable']).rename(columns={0: 'id', 'value': f"forecast_{i}"})
+                output_df = output_df.rename_axis('index').sort_values(by=['id', 'index']).reset_index(drop=True)
+                df_list.append(output_df)
         output_df = pd.concat(df_list, axis=1)
         output_df = output_df.loc[:,~output_df.columns.duplicated()]
         
@@ -1305,8 +1595,23 @@ def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quant
         
     # weights df merge with id
     stat_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), stat_wts_df], axis=1)
-    encoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), encoder_wts_df], axis=1)
-    decoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), decoder_wts_df], axis=1)
+
+    # v0.1.38
+    hist_periods = pd.DataFrame(np.arange(hist_len).reshape(-1, )).rename(columns={0: 'period'})
+    hist_periods = pd.concat([hist_periods] * hist_len, ignore_index=True)
+    encoder_wts_df = pd.concat([hist_periods, encoder_wts_df], axis=1)
+    hid_df = pd.DataFrame(np.repeat(id_arr.reshape(-1, 1), hist_len, axis=0))
+    hid_df.columns = ['id']
+    encoder_wts_df = pd.concat([hid_df, encoder_wts_df], axis=1)
+    #encoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), encoder_wts_df], axis=1)
+
+    # v0.1.38
+    decoder_wts_df = pd.concat([date_df, decoder_wts_df], axis=1)
+    fid_df = pd.DataFrame(np.repeat(id_arr.reshape(-1, 1), f_len, axis=0))
+    fid_df.columns = ['id']
+    decoder_wts_df = pd.concat([fid_df, decoder_wts_df], axis=1)
+    #decoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), decoder_wts_df], axis=1)
+
     print(stat_wts_df.shape, encoder_wts_df.shape, decoder_wts_df.shape)
         
     return forecast_df, stat_wts_df, encoder_wts_df, decoder_wts_df
@@ -1357,23 +1662,38 @@ class Temporal_Fusion_Transformer:
               train_dataset, 
               test_dataset,
               loss_function, 
-              metric, 
-              learning_rate,
-              max_epochs, 
-              min_epochs,
-              train_steps_per_epoch,
-              test_steps_per_epoch,
-              patience,
-              weighted_training,
-              model_prefix,
-              logdir,
+              metric='MSE',
+              learning_rate=0.0001,
+              max_epochs=100,
+              min_epochs=10,
+              prefill_buffers=False,
+              num_train_samples=200000,
+              num_test_samples=50000,
+              train_batch_size=64,
+              train_steps_per_epoch=200,
+              test_steps_per_epoch=100,
+              patience=10,
+              weighted_training=False,
+              model_prefix='./tft_model',
+              logdir='/tmp/tft_logs',
+              load_model=None,
               opt=None,
               clipnorm=None):
-        
-        # Initialize Weights
-        for x,y,s,w in train_dataset.take(1):
-            self.model(x, training=False)
-        
+
+        if load_model is None:
+            # Initialize Weights
+            for x,y,s,w in train_dataset.take(1):
+                self.model(x, training=False)
+        else:
+            # Initialize Weights
+            for x, y, s, w in train_dataset.take(1):
+                self.model(x, training=False)
+            saved_model = tf.keras.models.load_model(load_model)
+            self.model.set_weights(saved_model.get_weights())
+            del saved_model
+            gc.collect()
+            print("Saved model: {} loaded. Continuing training ...".format(load_model))
+
         best_model = TFT_Train(self.model, 
                                train_dataset, 
                                test_dataset, 
@@ -1383,6 +1703,10 @@ class Temporal_Fusion_Transformer:
                                learning_rate,
                                max_epochs, 
                                min_epochs,
+                               prefill_buffers,
+                               num_train_samples,
+                               num_test_samples,
+                               train_batch_size,
                                train_steps_per_epoch,
                                test_steps_per_epoch,
                                patience,
