@@ -1,6 +1,9 @@
 #!/usr/bin/env python
 # coding: utf-8
 
+# In[1]:
+
+
 import numpy as np
 import math as m
 import tensorflow as tf
@@ -12,7 +15,6 @@ import absl.logging
 absl.logging.set_verbosity(absl.logging.ERROR)
 import pandas as pd
 
-
 # Distribution Sampling functions
 
 # negbin
@@ -64,7 +66,7 @@ def GumbelSample(a, b, n_samples=1):
 
 
 # Self-Attention
-def ScaledDotProductAttention(q, k, v, causal_mask):
+def ScaledDotProductAttention(q, k, v, causal_mask, padding_mask):
     matmul_qk = tf.matmul(q, k, transpose_b=True)  # (..., seq_len_q, seq_len_k)
     dk = tf.cast(tf.shape(k)[-1], tf.float32)
     scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)
@@ -72,6 +74,9 @@ def ScaledDotProductAttention(q, k, v, causal_mask):
     # add the mask to the scaled tensor.
     if causal_mask is not None:
         scaled_attention_logits += (causal_mask * -1e9)
+        
+    if padding_mask is not None:
+        scaled_attention_logits += (padding_mask * -1e9)
     
     # softmax is normalized on the last axis (seq_len_k) so that the scores add up to 1.
     attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)  # (..., seq_len_q, seq_len_k)
@@ -112,7 +117,7 @@ class TFTMultiHeadAttention(tf.keras.layers.Layer):
 
         self.w_o = tf.keras.layers.Dense(units=d_model, use_bias=False)
 
-    def call(self, q, k, v, causal_mask, training):
+    def call(self, q, k, v, causal_mask, padding_mask, training):
         """
         q: Query tensor of shape=(?, T, d_model)
         k: Key of shape=(?, T, d_model)
@@ -129,7 +134,7 @@ class TFTMultiHeadAttention(tf.keras.layers.Layer):
             qs = self.qs_layers[i](q)
             ks = self.ks_layers[i](k)
             vs = self.vs_layers[i](v)
-            head, attn = ScaledDotProductAttention(qs, ks, vs, causal_mask)
+            head, attn = ScaledDotProductAttention(qs, ks, vs, causal_mask, padding_mask)
             head_dropout = tf.keras.layers.Dropout(self.dropout_rate)(head)
             heads.append(head_dropout)
             attns.append(attn)
@@ -512,9 +517,9 @@ class Attention_Layer(tf.keras.layers.Layer):
                                            activation=None)
         self.add_norm = tft_add_and_norm_layer()
 
-    def call(self, x, causal_mask, training):
+    def call(self, x, causal_mask, padding_mask, training):
 
-        attn_out, _ = self.mha(x, x, x, causal_mask, training=training) # (q,k,v,mask,training)
+        attn_out, _ = self.mha(x, x, x, causal_mask, padding_mask, training=training) # (q,k,v,mask,training)
 
         # gating layer
         attn_out, _ = self.gate(attn_out, training=training)
@@ -538,11 +543,11 @@ class Attention_Stack(tf.keras.layers.Layer):
                                        additional_context=False,
                                        return_gate=False)
 
-    def call(self, x, causal_mask, training):
+    def call(self, x, causal_mask, padding_mask, training):
 
         attn_out = x
         for i in range(self.num_layers):
-            attn_out = self.attn_layers[i](attn_out, causal_mask, training=training)
+            attn_out = self.attn_layers[i](attn_out, causal_mask, padding_mask, training=training)
 
         # final GRN layer
         attn_out = self.grn_final(attn_out, training=training)
@@ -571,7 +576,6 @@ class final_gating_layer(tf.keras.layers.Layer):
 
         return out
 
-
 # TFT Model
 
 class TFT(tf.keras.Model):
@@ -691,7 +695,7 @@ class TFT(tf.keras.Model):
         mask = causal_mask(self.hist_len + self.f_len)
 
         # Attention stack
-        attn_out = self.self_attention(enriched_features, mask, training=training)
+        attn_out = self.self_attention(enriched_features, mask, padding_mask, training=training)
 
         # pff
         attn_out = self.pff_layer([attn_out, temporal_features], training=training)
@@ -736,7 +740,6 @@ class TFT(tf.keras.Model):
             return out, parameters, stat_weights, past_weights, future_weights
             
 
-
 # TFT Wrapper
 
 class TFT_Model(tf.keras.Model):
@@ -750,6 +753,7 @@ class TFT_Model(tf.keras.Model):
                  max_inp_len,
                  loss_type,
                  num_quantiles=1,
+                 decoder_start_tokens=4,
                  dropout_rate=0.1):
 
         super(TFT_Model, self).__init__()
@@ -766,6 +770,7 @@ class TFT_Model(tf.keras.Model):
         self.max_inp_len = max_inp_len
         self.num_quantiles = num_quantiles
         self.dropout_rate = dropout_rate
+        self.start_token_length = decoder_start_tokens
         
         if (len(self.col_index_dict.get('static_num_indices')[0])==0) and (len(self.col_index_dict.get('static_cat_indices')[0])==0):
             self.static_variables = False
@@ -855,12 +860,12 @@ class TFT_Model(tf.keras.Model):
             for colname in self.stat_num_col_names:
                 self.stat_linear_transform_layers[colname] = tf.keras.layers.Dense(units=d_model, use_bias=False)
         
-        if len(self.known_num_col_names)>0:
+        if len(self.known_num_col_names)>=0:
             self.known_linear_transform_layers = {}
             for colname in self.known_num_col_names + ['rel_age']:
                 self.known_linear_transform_layers[colname] = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=d_model, use_bias=False))
         
-        if len(self.unknown_num_col_names)>0:
+        if len(self.unknown_num_col_names)>=0:
             self.unknown_linear_transform_layers = {}
             for colname in self.unknown_num_col_names + ['rel_age']:
                 self.unknown_linear_transform_layers[colname] = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(units=d_model, use_bias=False))
@@ -881,6 +886,15 @@ class TFT_Model(tf.keras.Model):
         encoder_vars_list = [target[:,:self.hist_len,:]]
         decoder_vars_list = []
         
+        # decoder start token - latest 'n' actuals before start of decoding
+        
+        start_token = target[:,self.hist_len-self.start_token_length:self.hist_len,:]
+        #B, _, _ = tf.shape(start_token)
+        start_token = tf.reshape(start_token, [-1,1,self.start_token_length*self.d_model])
+        start_token = tf.tile(start_token, [1,self.f_len,1])
+        decoder_vars_list.append(start_token)
+        future_cols_ordered_list  = future_cols_ordered_list + ["decoder_start_token"]
+        
         # static numeric
         if len(self.stat_num_indices)>0:
             stat_cols_ordered_list = stat_cols_ordered_list + self.stat_num_col_names
@@ -972,8 +986,13 @@ class TFT_Model(tf.keras.Model):
             o, s, s_wts, p_wts, f_wts = self.model([encoder_vars_list, decoder_vars_list, mask, scale], training=training)
         
         # Average feature weights across time dim
-        p_wts = tf.math.reduce_mean(p_wts, axis=1)
-        f_wts = tf.math.reduce_mean(f_wts, axis=1)
+        #p_wts = tf.math.reduce_mean(p_wts, axis=1)
+        #f_wts = tf.math.reduce_mean(f_wts, axis=1)
+
+        # Retain period-wise importance
+        bs = tf.shape(p_wts)[0]
+        p_wts = tf.reshape(p_wts, [bs * self.hist_len, -1])
+        f_wts = tf.reshape(f_wts, [bs * self.f_len, -1])
         
         return o, s, ([stat_cols_ordered_list,past_cols_ordered_list,future_cols_ordered_list], [s_wts, p_wts, f_wts])
     
@@ -987,6 +1006,10 @@ def TFT_Train(model,
               learning_rate,
               max_epochs, 
               min_epochs,
+              prefill_buffers,
+              num_train_samples,
+              num_test_samples,
+              train_batch_size,
               train_steps_per_epoch,
               test_steps_per_epoch,
               patience,
@@ -1125,91 +1148,228 @@ def TFT_Train(model,
     model_list = []
     best_model = None
     time_since_improvement = 0
-        
-    # train loop
-    for epoch in range(max_epochs):
-        print("Epoch {}/{}". format(epoch, max_epochs)) 
+    #####################################################################
+
+    if prefill_buffers:
+        print("prefetching training samples ... ")
+        # get batch size
+        batch_size = 0
+        for x, y, s, w in train_dataset.take(1):
+            batch_size = int(x.shape[0])
+
+        x_train = []
+        y_train = []
+        train_scale = []
+        train_wts = []
         for step, (x_batch, y_batch, scale, wts) in enumerate(train_dataset):
-            if step > train_steps_per_epoch:
+            x_train.append(x_batch)
+            y_train.append(y_batch)
+            train_scale.append(scale)
+            train_wts.append(wts)
+            if (step + 1) * batch_size >= num_train_samples:
                 break
-            else:
+
+        # concat
+        x_train = tf.concat(x_train, axis=0)
+        y_train = tf.concat(y_train, axis=0)
+        train_scale = tf.concat(train_scale, axis=0)
+        train_wts = tf.concat(train_wts, axis=0)
+        print("Training Samples Gathered: ", x_train.shape[0])
+
+        print("prefetching test samples ... ")
+        x_test = []
+        y_test = []
+        test_scale = []
+        test_wts = []
+        for step, (x_batch, y_batch, scale, wts) in enumerate(test_dataset):
+            x_test.append(x_batch)
+            y_test.append(y_batch)
+            test_scale.append(scale)
+            test_wts.append(wts)
+            if (step + 1) * batch_size >= num_test_samples:
+                break
+
+        # concat
+        x_test = tf.concat(x_test, axis=0)
+        y_test = tf.concat(y_test, axis=0)
+        test_scale = tf.concat(test_scale, axis=0)
+        test_wts = tf.concat(test_wts, axis=0)
+        print("Test Samples Gathered: ", x_test.shape[0])
+
+        num_train_batches = int(x_train.shape[0] // train_batch_size)
+        num_test_batches = int(x_test.shape[0] // train_batch_size)
+
+        for epoch in range(max_epochs):
+            print("Epoch {}/{}".format(epoch, max_epochs))
+            for i in range(num_train_batches):
+                x_batch = x_train[i * train_batch_size:(i + 1) * train_batch_size]
+                y_batch = y_train[i * train_batch_size:(i + 1) * train_batch_size]
+                scale = train_scale[i * train_batch_size:(i + 1) * train_batch_size]
+                wts = train_wts[i * train_batch_size:(i + 1) * train_batch_size]
                 train_loss, train_out = trainstep(model, optimizer, x_batch, y_batch, scale, wts, training=True)
                 out_len = tf.shape(train_out)[1]
                 train_loss_avg.update_state(train_loss)
-                if loss_type in ['Normal','Poisson','Negbin']:
-                    train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out)
-                elif loss_type in ['Point','Quantile']:
-                    train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out*scale[:,-out_len:,:])
+                if loss_type in ['Normal', 'Poisson', 'Negbin']:
+                    train_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :], train_out)
+                elif loss_type in ['Point', 'Quantile']:
+                    train_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :],
+                                              train_out * scale[:, -out_len:, :])
                 with train_summary_writer.as_default():
                     tf.summary.scalar('loss', train_loss_avg.result(), step=epoch)
                     tf.summary.scalar('accuracy', train_metric.result(), step=epoch)
-          
-        for step, (x_batch, y_batch, scale, wts) in enumerate(test_dataset):
-            if step > train_steps_per_epoch:
-                break
-            else:
-                test_loss, test_out = teststep(model, x_batch, y_batch, scale, wts, training=False)
+
+            for i in range(num_test_batches):
+                x_batch = x_test[i * train_batch_size:(i + 1) * train_batch_size]
+                y_batch = y_test[i * train_batch_size:(i + 1) * train_batch_size]
+                scale = test_scale[i * train_batch_size:(i + 1) * train_batch_size]
+                wts = test_wts[i * train_batch_size:(i + 1) * train_batch_size]
+                test_loss, test_out = trainstep(model, optimizer, x_batch, y_batch, scale, wts, training=False)
                 out_len = tf.shape(test_out)[1]
                 test_loss_avg.update_state(test_loss)
-                if loss_type in ['Normal','Poisson','Negbin']:
-                    test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out)
-                elif loss_type in ['Point','Quantile']:
-                    test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out*scale[:,-out_len:,:])
+                if loss_type in ['Normal', 'Poisson', 'Negbin']:
+                    test_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :], test_out)
+                elif loss_type in ['Point', 'Quantile']:
+                    test_metric.update_state(y_batch[:, -out_len:, :] * scale[:, -out_len:, :],
+                                             test_out * scale[:, -out_len:, :])
                 with test_summary_writer.as_default():
                     tf.summary.scalar('loss', test_loss_avg.result(), step=epoch)
                     tf.summary.scalar('accuracy', test_metric.result(), step=epoch)
-              
-        print("Epoch: {}, train_loss: {}, test_loss: {}, train_metric: {}, test_metric: {}".format(epoch, 
-                                                                                                    train_loss_avg.result().numpy(),
-                                                                                                    test_loss_avg.result().numpy(),
-                                                                                                    train_metric.result().numpy(),
-                                                                                                    test_metric.result().numpy()))
-    
-        # record losses & metric in lists
-        train_loss_results.append(train_loss_avg.result().numpy())
-        train_metric_results.append(train_metric.result().numpy())
-        test_loss_results.append(test_loss_avg.result().numpy())
-        test_metric_results.append(test_metric.result().numpy())
-            
-        # reset states
-        train_loss_avg.reset_states()
-        train_metric.reset_states()
-        test_loss_avg.reset_states()
-        test_metric.reset_states()
-        
-        # Save Model
-        model_path = model_prefix + '_' + str(epoch) 
-        model_list.append(model_path)
-        
-        # track & save best model
-        if test_loss_results[epoch]==np.min(test_loss_results):
-            best_model = model_path
-            tf.keras.models.save_model(model, model_path)
-            # reset time_since_improvement
-            time_since_improvement = 0
-        else:
-            time_since_improvement += 1
-            
-        model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
-        print("Best Model: ", best_model)
-            
-        # remove older models
-        if len(model_list)>patience:
-            for m in model_list[:-patience]:
-                if m != best_model:
-                    try:
-                        shutil.rmtree(m)
-                    except:
-                        pass
-                
-        if (time_since_improvement > patience) and (epoch > min_epochs):
-            print("Terminating Training. Best Model path: {}".format(best_model))
-            model_tracker_file.close()
-            break
-            
-        # write after each epoch
-        model_tracker_file.flush()
-        
+
+            print("Epoch: {}, train_loss: {}, test_loss: {}, train_metric: {}, test_metric: {}".format(epoch,
+                                                                                                       train_loss_avg.result().numpy(),
+                                                                                                       test_loss_avg.result().numpy(),
+                                                                                                       train_metric.result().numpy(),
+                                                                                                       test_metric.result().numpy()))
+
+            # record losses & metric in lists
+            train_loss_results.append(train_loss_avg.result().numpy())
+            train_metric_results.append(train_metric.result().numpy())
+            test_loss_results.append(test_loss_avg.result().numpy())
+            test_metric_results.append(test_metric.result().numpy())
+
+            # reset states
+            train_loss_avg.reset_states()
+            train_metric.reset_states()
+            test_loss_avg.reset_states()
+            test_metric.reset_states()
+
+            # Save Model
+            model_path = model_prefix + '_' + str(epoch)
+            model_list.append(model_path)
+
+            # track & save best model
+            if test_loss_results[epoch] == np.min(test_loss_results):
+                best_model = model_path
+                tf.keras.models.save_model(model, model_path)
+                # reset time_since_improvement
+                time_since_improvement = 0
+            else:
+                time_since_improvement += 1
+
+            model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
+            print("Best Model: ", best_model)
+
+            # remove older models
+            if len(model_list) > patience:
+                for m in model_list[:-patience]:
+                    if m != best_model:
+                        try:
+                            shutil.rmtree(m)
+                        except:
+                            pass
+
+            if (time_since_improvement > patience) and (epoch > min_epochs):
+                print("Terminating Training. Best Model path: {}".format(best_model))
+                model_tracker_file.close()
+                break
+
+            # write after each epoch
+            model_tracker_file.flush()
+    else:
+        # train loop
+        for epoch in range(max_epochs):
+            print("Epoch {}/{}". format(epoch, max_epochs))
+            for step, (x_batch, y_batch, scale, wts) in enumerate(train_dataset):
+                if step > train_steps_per_epoch:
+                    break
+                else:
+                    train_loss, train_out = trainstep(model, optimizer, x_batch, y_batch, scale, wts, training=True)
+                    out_len = tf.shape(train_out)[1]
+                    train_loss_avg.update_state(train_loss)
+                    if loss_type in ['Normal','Poisson','Negbin']:
+                        train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out)
+                    elif loss_type in ['Point','Quantile']:
+                        train_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], train_out*scale[:,-out_len:,:])
+                    with train_summary_writer.as_default():
+                        tf.summary.scalar('loss', train_loss_avg.result(), step=epoch)
+                        tf.summary.scalar('accuracy', train_metric.result(), step=epoch)
+
+            for step, (x_batch, y_batch, scale, wts) in enumerate(test_dataset):
+                if step > train_steps_per_epoch:
+                    break
+                else:
+                    test_loss, test_out = teststep(model, x_batch, y_batch, scale, wts, training=False)
+                    out_len = tf.shape(test_out)[1]
+                    test_loss_avg.update_state(test_loss)
+                    if loss_type in ['Normal','Poisson','Negbin']:
+                        test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out)
+                    elif loss_type in ['Point','Quantile']:
+                        test_metric.update_state(y_batch[:,-out_len:,:]*scale[:,-out_len:,:], test_out*scale[:,-out_len:,:])
+                    with test_summary_writer.as_default():
+                        tf.summary.scalar('loss', test_loss_avg.result(), step=epoch)
+                        tf.summary.scalar('accuracy', test_metric.result(), step=epoch)
+
+            print("Epoch: {}, train_loss: {}, test_loss: {}, train_metric: {}, test_metric: {}".format(epoch,
+                                                                                                        train_loss_avg.result().numpy(),
+                                                                                                        test_loss_avg.result().numpy(),
+                                                                                                        train_metric.result().numpy(),
+                                                                                                        test_metric.result().numpy()))
+
+            # record losses & metric in lists
+            train_loss_results.append(train_loss_avg.result().numpy())
+            train_metric_results.append(train_metric.result().numpy())
+            test_loss_results.append(test_loss_avg.result().numpy())
+            test_metric_results.append(test_metric.result().numpy())
+
+            # reset states
+            train_loss_avg.reset_states()
+            train_metric.reset_states()
+            test_loss_avg.reset_states()
+            test_metric.reset_states()
+
+            # Save Model
+            model_path = model_prefix + '_' + str(epoch)
+            model_list.append(model_path)
+
+            # track & save best model
+            if test_loss_results[epoch]==np.min(test_loss_results):
+                best_model = model_path
+                tf.keras.models.save_model(model, model_path)
+                # reset time_since_improvement
+                time_since_improvement = 0
+            else:
+                time_since_improvement += 1
+
+            model_tracker_file.write('best_model path after epochs ' + str(epoch) + ': ' + best_model + '\n')
+            print("Best Model: ", best_model)
+
+            # remove older models
+            if len(model_list)>patience:
+                for m in model_list[:-patience]:
+                    if m != best_model:
+                        try:
+                            shutil.rmtree(m)
+                        except:
+                            pass
+
+            if (time_since_improvement > patience) and (epoch > min_epochs):
+                print("Terminating Training. Best Model path: {}".format(best_model))
+                model_tracker_file.close()
+                break
+
+            # write after each epoch
+            model_tracker_file.flush()
+
     return best_model
         
 def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quantiles):
@@ -1240,11 +1400,11 @@ def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quant
     encoder_columns_string = []
     decoder_columns_string = []
     for col in stat_columns:
-        stat_columns_string.append(col) 
+        stat_columns_string.append(col.numpy().decode("utf-8")) 
     for col in encoder_columns:
-        encoder_columns_string.append(col)
+        encoder_columns_string.append(col.numpy().decode("utf-8"))
     for col in decoder_columns:
-        decoder_columns_string.append(col)
+        decoder_columns_string.append(col.numpy().decode("utf-8"))
                 
     stat_wts, encoder_wts, decoder_wts = wts_list
 
@@ -1291,8 +1451,23 @@ def TFT_Infer(model, inputs, loss_type, hist_len, f_len, target_index, num_quant
         
     # weights df merge with id
     stat_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), stat_wts_df], axis=1)
-    encoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), encoder_wts_df], axis=1)
-    decoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), decoder_wts_df], axis=1)
+
+    # v0.1.38
+    hist_periods = pd.DataFrame(np.arange(hist_len).reshape(-1, )).rename(columns={0: 'period'})
+    hist_periods = pd.concat([hist_periods] * hist_len, ignore_index=True)
+    encoder_wts_df = pd.concat([hist_periods, encoder_wts_df], axis=1)
+    hid_df = pd.DataFrame(np.repeat(id_arr.reshape(-1, 1), hist_len, axis=0))
+    hid_df.columns = ['id']
+    encoder_wts_df = pd.concat([hid_df, encoder_wts_df], axis=1)
+    #encoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), encoder_wts_df], axis=1)
+
+    # v0.1.38
+    decoder_wts_df = pd.concat([date_df, decoder_wts_df], axis=1)
+    fid_df = pd.DataFrame(np.repeat(id_arr.reshape(-1, 1), f_len, axis=0))
+    fid_df.columns = ['id']
+    decoder_wts_df = pd.concat([fid_df, decoder_wts_df], axis=1)
+    #decoder_wts_df = pd.concat([pd.DataFrame(id_arr.reshape(-1,1)), decoder_wts_df], axis=1)
+
     print(stat_wts_df.shape, encoder_wts_df.shape, decoder_wts_df.shape)
         
     return forecast_df, stat_wts_df, encoder_wts_df, decoder_wts_df
@@ -1309,6 +1484,7 @@ class Temporal_Fusion_Transformer:
                  max_inp_len,
                  loss_type,
                  num_quantiles=1,
+                 decoder_start_tokens=4,
                  dropout_rate=0.1):
         
         self.col_index_dict = col_index_dict
@@ -1321,6 +1497,7 @@ class Temporal_Fusion_Transformer:
         self.loss_type = loss_type
         self.num_quantiles = num_quantiles
         self.dropout_rate = dropout_rate
+        self.decoder_start_tokens = decoder_start_tokens
         self.target_col_name, self.target_index = self.col_index_dict.get('target_index')
     
     def build(self):
@@ -1334,22 +1511,27 @@ class Temporal_Fusion_Transformer:
                                   self.max_inp_len,
                                   self.loss_type,
                                   self.num_quantiles,
+                                  self.decoder_start_tokens,
                                   self.dropout_rate)
         
     def train(self, 
               train_dataset, 
               test_dataset,
               loss_function, 
-              metric, 
-              learning_rate,
-              max_epochs, 
-              min_epochs,
-              train_steps_per_epoch,
-              test_steps_per_epoch,
-              patience,
-              weighted_training,
-              model_prefix,
-              logdir,
+              metric='MSE',
+              learning_rate=0.0001,
+              max_epochs=100,
+              min_epochs=10,
+              prefill_buffers=False,
+              num_train_samples=200000,
+              num_test_samples=50000,
+              train_batch_size=64,
+              train_steps_per_epoch=200,
+              test_steps_per_epoch=100,
+              patience=10,
+              weighted_training=False,
+              model_prefix='./tft_model',
+              logdir='/tmp/tft_logs',
               opt=None,
               clipnorm=None):
         
@@ -1366,6 +1548,10 @@ class Temporal_Fusion_Transformer:
                                learning_rate,
                                max_epochs, 
                                min_epochs,
+                               prefill_buffers,
+                               num_train_samples,
+                               num_test_samples,
+                               train_batch_size,
                                train_steps_per_epoch,
                                test_steps_per_epoch,
                                patience,
@@ -1384,4 +1570,4 @@ class Temporal_Fusion_Transformer:
         forecast, stat_wts_df, encoder_wts_df, decoder_wts_df = TFT_Infer(self.model, inputs, self.loss_type, self.max_inp_len, self.forecast_horizon, self.target_index, self.num_quantiles)
         
         return forecast, [stat_wts_df, encoder_wts_df, decoder_wts_df]
-       
+
